/*
 * Copyright © 2017-2022 NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of Nvidia Corporation and its affiliates
 * (the "Company") and all right, title, and interest in and to the software
 * product, including all associated intellectual property rights, are and
 * shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 */

#ifndef SRC_RIVERMAX_EX_API_H_
#define SRC_RIVERMAX_EX_API_H_

#include <rivermax_api.h>

#ifdef __linux__
#include <netinet/in.h>
#define __export
#else
#include <winsock2.h>
#ifdef _USRDLL
#define __export __declspec(dllexport)
#else
#define __export __declspec(dllimport)
#endif
#endif

#if __cplusplus < 201103L
#include <stdint.h>
#else
#include <cinttypes>
#endif

/**
 * @brief: Shift for flags from the legacy rivermax_api.h.
 */
#define EX_API_SHIFT_SIZE 20
#define EX_API_SHIFT(BIT) ((1ULL << EX_API_SHIFT_SIZE) << BIT)
/**
 * @brief: Rivermax device configuration flags.
 *
 */
typedef enum rmax_dev_config_flags_ex {
    RMAX_DEV_CONFIG_EX_NONE = 0x00ULL,
    /**
     * Enables eCPRI protocol usage.
     *
     * Use @ref rmax_device_config_t.ip_address to set the IP address of the device
     * this protocol will be enabled on.
     */
    RMAX_DEV_CONFIG_EX_ECPRI_PROTOCOL_CONFIG = EX_API_SHIFT(0),
} rmax_dev_config_flags_ex_t;

/**
 * @brief: Bit mask which user provide to query needed device capabilities
 *
 * RMAX_CAP_NONE - query nothing capabilities
 * RMAX_CAP_EX_STREAM_ECPRI_SYMBOL_PRB_PLACEMENT_ORDER - stream which locate the incoming
 * packets according to eCPRI protocol is supported
 */
typedef enum rmax_device_caps_mask_ex {
    RMAX_CAP_EX_NONE                                    = 0,
    RMAX_CAP_EX_STREAM_ECPRI_SYMBOL_PRB_PLACEMENT_ORDER = EX_API_SHIFT(3),
} rmax_device_caps_mask_ex_t;


/**
 * @brief: Rivermax in buffer attributes flags.
 *
 * RMAX_IN_BUFFER_ATTER_EX_STREAM_ECPRI_SYMBOL_PRB_PLACEMENT_ORDER - when set input stream will
 * locate the incoming packets according to eCPRI protocol in the following way:
 * Packet index = symbolId[4b]startPrbu[9b] <==> Packet index = (symbolId & 0xF) << 9 | (startPrbu & 0x1FF)
 */
typedef enum rmax_in_buffer_attr_flags_ex {
    RMAX_IN_BUFFER_ATTER_EX_FLAG_NONE = 0x00ULL,
    RMAX_IN_BUFFER_ATTER_EX_STREAM_ECPRI_SYMBOL_PRB_PLACEMENT_ORDER = EX_API_SHIFT(0),
} rmax_in_buffer_attr_flags_ex_t;

/**
 * @brief: Rivermax network flow layer 2 attributes.
 *
 * @param [in] source_mac - Source MAC address.
 * @param [in] destination_mac - Destination MAC address.
 * @param [in] vlan_id - VLAN identifier.
 */
typedef struct rmax_in_flow_attr_l2_ex {
    uint64_t source_mac : 48;
    uint64_t destination_mac : 48;
    uint64_t vlan_id : 12;
} rmax_in_flow_attr_l2_ex_t;

/**
 * @brief: Rivermax network flow layer 3 attributes.
 *
 * @param [in] source_ip - Source IP address.
 * @param [in] destination_ip - Destination IP address.
 */
typedef struct rmax_in_flow_attr_l3_ex {
    uint32_t source_ip;
    uint32_t destination_ip;
} rmax_in_flow_attr_l3_ex_t;

/**
 * @brief: Rivermax network flow layer 4 attributes.
 *
 * @param [in] source_port - Source port address.
 * @param [in] destination_port - Destination port address.
 */
typedef struct rmax_in_flow_attr_l4_ex {
    uint16_t source_port;
    uint16_t destination_port;
} rmax_in_flow_attr_l4_ex_t;

/**
 * @brief: Rivermax network flow eCPRI application layer attributes.
 *
 * @param [in] pc_id - IQ data transfer message series identifier.
 * @param [in] subframe_id - Sub-frame identifier.
 * @param [in] slot_id - Slot identifier.
 * @param [in] section_id - Section identifier.
 */
typedef struct rmax_in_flow_attr_app_ecpri_ex {
    uint64_t pc_id : 16;
    uint64_t subframe_id : 4;
    uint64_t slot_id : 6;
    uint64_t section_id : 12;
} rmax_in_flow_attr_app_ecpri_ex_t;

/**
 * @brief: Rivermax in flow type flags.
 */
typedef enum rmax_in_flow_app_layer_ex {
    RMAX_IN_FLOW_APP_LAYER_EX_NONE, /**< There is no specific application flow type */
    RMAX_IN_FLOW_APP_LAYER_EX_ECPRI, /**< eCPRI flow type */
} rmax_in_flow_app_layer_ex_t;

/**
 * @brief: Rivermax network flow.
 *
 * Represents a network flow defined by L2/L3/L4 and potential application layer.
 * Single or multiple network flows can be read from a single input stream.
 * After input stream creation application defines which incoming network flows the NIC
 * will deliver to input stream via @ref rmax_in_attach_flow_ex(flow_attr) - aka application attaches flows
 * to input stream.
 * Flows can be dynamically added to input stream or removed from input stream.
 * In case no networks flows are attached to input stream no traffic will be delivered to it.
 *
 * Each layer attributes consists of mask and value. In order to much the received flow on the
 * needed value, the mask should be enabled and the value should be set to the required much value.
 *
 * @param [in] l2_mask - Mask for layer 2 attributes, see @ref rmax_in_flow_attr_l2_ex_t.
 * @param [in] l2_value - Value for layer 2 attributes, see @ref rmax_in_flow_attr_l2_ex_t.
 * @param [in] l3_mask - Mask for layer 3 attributes, see @ref rmax_in_flow_attr_l3_ex_t.
 * @param [in] l3_value - Value for layer 3 attributes, see @ref rmax_in_flow_attr_l3_ex_t.
 * @param [in] l4_mask - Mask for layer 4 attributes, see @ref rmax_in_flow_attr_l4_ex_t.
 * @param [in] l4_value - Value for layer 4 attributes, see @ref rmax_in_flow_attr_l4_ex_t.
 * @param [in] app_layer_type - Type of application layer, see @ref rmax_in_flow_app_layer_ex_t.
 * @param [in] app_layer - Application layer specific attributes and values.
 *
 * @param [in] flow_id - Flow ID is non-zero value, set by application and is returned by input stream upon
 *                       packet reception via @rmax_in_packet_info field of @rmax_in_completion.
 *                       Allows to dispatch the received packet to the right flow (without parsing headers),
 *                       this is essential in case multiple flows are associated with the same input stream.
 */
typedef struct rmax_in_flow_attr_ex {
    struct {
        rmax_in_flow_attr_l2_ex_t mask;
        rmax_in_flow_attr_l2_ex_t value;
    } l2;
    struct {
        rmax_in_flow_attr_l3_ex_t mask;
        rmax_in_flow_attr_l3_ex_t value;
    } l3;
    struct {
        rmax_in_flow_attr_l4_ex_t mask;
        rmax_in_flow_attr_l4_ex_t value;
    } l4;

    rmax_in_flow_app_layer_ex_t app_layer_type;
    union {
        struct {
            rmax_in_flow_attr_app_ecpri_ex_t mask;
            rmax_in_flow_attr_app_ecpri_ex_t value;
        } ecpri;
    } app_layer;

    uint32_t flow_id;
} rmax_in_flow_attr_ex_t;

#ifdef __cplusplus
extern "C" {
#endif

/* General use functions */

/* Input Stream functions */

/**
 * TODO: Remove the commented comments below, when the below API is verified.
 */

/**
 * @brief return the memory size that matched the given parameters
 * This function can be used to calculate the memory size needed by Rivermax to create
 * the rx session. After getting the size the user can allocate this amount and pass the pointer
 * to Rivermax using the @ref rmax_in_create_stream function in the @ref buffer_attr struct
 * @param [in] rx_type - the stream type
 * @param [in] flags - as described in rmax_in_flags
 * @param [in] local_nic_addr - IP address of the local NIC. Rivermax uses @local_nic_addr
 *     in order to resolve the local NIC.
 * @param [in] buffer_attr - the buffer parameters
 * @param [out] payload_size - the payload buffer size to use
 * @param [out] header_size - if not NULL Rivermax will return the size of the user header buffer
 *     to allocate. Otherwise, the returned payload_size size will be large enough to include both
 *     header and payload sizes.
 *     Rivermax will know to use this size if user provides a pointer in hdr_ptr or the 
 *     @ref buffer_attr when creating the session.
 *     In case of rx_type = @ref rmax_in_stream_type.RMAX_APP_PROTOCOL_PAYLOAD, it should be
 *     set to NULL, the return payload_size in this case will include only payload size.
 * @return status code as defined by @ref rmax_status_t
 * @note the pointer given in @ref rmax_in_create_stream must be aligned to cache line unless
 *     - RMAX_IN_BUFFER_ATTER_STREAM_RTP_SEQN_PLACEMENT_ORDER
 *     - RMAX_IN_BUFFER_ATTER_STREAM_RTP_EXT_SEQN_PLACEMENT_ORDER
 *     - RMAX_IN_BUFFER_ATTER_EX_STREAM_ECPRI_SYMBOL_PRB_PLACEMENT_ORDER // XXX remove from public documentation and add to internal h file
 *     flags are being used, in which case memory must be aligned to page size, see @rmax_in_buffer_attr_flags
 */
/** @note: eCPRI implementation notes:
 *
 *-------------------------------------------------------------------------------------------------------------------------
 * @param [in] rx_type - Need to enforce for DPP:
 *                       (rx_type == @ref rmax_in_stream_type.RMAX_APP_PROTOCOL_PACKET) ||
 *                       (rx_type == @ref rmax_in_stream_type.RMAX_APP_PROTOCOL_PAYLOAD)
 *                       In case of eCPRI, we are adding a different behavior than the
 *                       current behavior of @ref rmax_in_stream_type.RMAX_APP_PROTOCOL_PACKET flag.
 *                       This flag says that the application will get L4 payload, which is not true for current
 *                       packet eCPRI, which is only the payload of the eCPRI packet on top of L2.
 *                       Thus, for eCPRI, need to enforce:
 *                       rx_type == @ref rmax_in_stream_type.RMAX_APP_PROTOCOL_PAYLOAD
 *-------------------------------------------------------------------------------------------------------------------------
 * @param [in] local_nic_addr - eCPRI currently is over L2, but the application will supply local NIC address
 *                              for NIC binding.
 *-------------------------------------------------------------------------------------------------------------------------
 * @param [in] buffer_attr - The buffer parameters:
 *          @param num_of_elements [in/out] - Application needs to set the number of PRBs in a slot (273 * 14),
 *                                            due to DPP scattering limitation, it should set (512 * 14)
 *                                            Rivermax will rounded up to the closest power of 2 (8192).
 *                                            TODO: Rethink for the GA.
 *          @param data [in] - The payload memory block parameters:
 *                   @param min_size [in] = max_size [in] = PRB size (48[B]).
 *                   @param ptr [in] - NULL/GPU allocation, then use new registration API.
 *                   @param stride_size [out] - Will be set to PRB size (48[B]) by Rivermax.
 *          @param hdr [in] - The header memory block parameters:
 *                   @param min_size [in] = max_size [in] = eCPRI header size (16[B]).
 *                   @param ptr [in] - NULL, the header will be removed by the DPP RQ.
 *                   @param stride_size [out] - Will be set to 0 by Rivermax.
 *          @param attr_flags [in] - Set
 *                   @ref rmax_ecpri_in_flow_attr_ex_t_t.RMAX_IN_BUFFER_ATTER_EX_STREAM_ECPRI_SYMBOL_PRB_PLACEMENT_ORDER.
 *-------------------------------------------------------------------------------------------------------------------------
 * @param [out] payload_size - The payload buffer size to use, will be set by Rivermax.
 *-------------------------------------------------------------------------------------------------------------------------
 * @param [out] header_size - Should be set to NULL by the user.
 *-------------------------------------------------------------------------------------------------------------------------
*/
//__export
//rmax_status_t rmax_in_query_buffer_size(rmax_in_stream_type rx_type,
//                                        struct sockaddr_in *local_nic_addr,
//                                        struct rmax_in_buffer_attr *buffer_attr,
//                                        size_t *payload_size, size_t *header_size);
/**
 * @brief Creates an input stream.
 * @param [in] type - input stream type, see @ref rmax_in_stream_type
 * @param [in] local_nic_addr - IP address of the local NIC. Rivermax uses @local_nic_addr in order
 *     to resolve the local NIC. Input stream is configured on that NIC.
 * @param [in] buffer_attr - buffer attributes as described in @ref rmax_in_buffer_attr
 * @param [in] timestamp_format - the timestamp format as described in @ref rmax_in_timestamp_format
 * @param [in] flags - as described in rmax_in_flags
 * @stream_id [out] - output argument that returns rmax_stream_id to use for this
 *   stream operations
 * @return status code as defined by @ref rmax_status_t
 */
 /** @note: eCPRI implementation notes:
  *
  *-------------------------------------------------------------------------------------------------------------------------
  * @param [in] type - Should be set to @ref rmax_in_stream_type.RMAX_APP_PROTOCOL_PAYLOAD.
  *-------------------------------------------------------------------------------------------------------------------------
  * @param [in] local_nic_addr - eCPRI currently is over L2, but the application will supply local NIC address
  *                              for NIC binding.
 *-------------------------------------------------------------------------------------------------------------------------
 * @param [in] buffer_attr - The buffer parameters:
 *          @param num_of_elements [in] - Application will set following the call to @ref rmax_in_query_buffer_size.
 *          @param data [in] - The payload memory block parameters:
 *                   @param min_size [in] = max_size [in] = PRB size (48[B]).
 *                   @param ptr [in] - GPU allocated memory pointer, could use new registration API.
 *                   @param stride_size [out] - Will be set to PRB size (48[B]) by Rivermax.
 *          @param hdr [in] - The header memory block parameters:
 *                   @param min_size [in] = max_size [in] = eCPRI header size (16[B]).
 *                   @param ptr [in] - NULL, the header will be removed by the DPP RQ.
 *                   @param stride_size [out] - Will be set to 0 by Rivermax.
 *          @param attr_flags [in] - Set
 *                   @ref rmax_in_buffer_attr_flags_ex_t.RMAX_IN_BUFFER_ATTER_EX_STREAM_ECPRI_SYMBOL_PRB_PLACEMENT_ORDER.
 *-------------------------------------------------------------------------------------------------------------------------
 * @param [in] timestamp_format - the timestamp format as described in @ref rmax_in_timestamp_format.
 *-------------------------------------------------------------------------------------------------------------------------
 * @param [in] flags - Set @ref rmax_in_flags.RMAX_IN_CREATE_STREAM_INFO_PER_PACKET.
 *-------------------------------------------------------------------------------------------------------------------------
 * @param stream_id [out] - output argument that returns rmax_stream_id to use for this stream operations.
 *-------------------------------------------------------------------------------------------------------------------------
 */
//__export
//rmax_status_t rmax_in_create_stream(rmax_in_stream_type rx_type, struct sockaddr_in *local_nic_addr,
//                                    struct rmax_in_buffer_attr *buffer_attr,
//                                    rmax_in_timestamp_format timestamp_format, rmax_in_flags flags,
//                                    rmax_stream_id *stream_id);

/**
 * @brief: Attach network flow to a Rivermax stream, experimental.
 *
 * @param [in] id - The id given in @ref rmax_in_create_stream.
 * @param [in] flow_attr - Network flow attributes see @rmax_in_flow_attr_ex_t.
 *
 * @note: This API is partially supported, should be used only for eCPRI use case.
 *
 * @return: Status code as defined by @ref rmax_status_t.
 */
__export
 rmax_status_t rmax_in_attach_flow_ex(rmax_stream_id id, rmax_in_flow_attr_ex_t* flow_attr);

/**
 * @brief: Detach network from a Rivermax stream, experimental.
 *
 * @param [in] id - The id given in @ref rmax_in_create_stream.
 * @param [in] flow_attr - Network flow attributes see @rmax_in_flow_attr_ex_t.
 *
 * @note: This API is partially supported, should be used only for eCPRI use case.
 *
 * @return: Status code as defined by @ref rmax_status_t.
 */
__export
rmax_status_t rmax_in_detach_flow_ex(rmax_stream_id id, rmax_in_flow_attr_ex_t* flow_attr);

#ifdef __cplusplus
}
#endif

#endif /* SRC_RIVERMAX_EX_API_H_ */

